import math
import time

import numpy as np
import pandas as pd
from pysat.examples.rc2 import RC2
from pysat.formula import WCNF

import trisicell as tsc


def phiscsb(df_input, alpha, beta, experiment=False):
    """Solving using PhISCS-B (only SC).

    a combinatorial approach for subperfect tumor phylogeny reconstruction
    via integrative use of single-cell and bulk sequencing data :cite:`PhISCS`.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1), absence (0) and missing entires (3).
    alpha : :obj:`float`
        False positive error rate.
    beta : :obj:`float`
        False negative error rate.
    experiment : :obj:`bool`, optional
        Is in the experiment mode (the log won't be shown), by default False

    Returns
    -------
    :class:`pandas.DataFrame`
        A conflict-free matrix in which rows are cells and columns are mutations.
        Values inside this matrix show the presence (1) and absence (0).
    """

    if not experiment:
        tsc.logg.info(f"running PhISCS-B with alpha={alpha}, beta={beta}")
    cells = list(df_input.index)
    snvs = list(df_input.columns)
    df_input = df_input.replace("?", 3)
    df_input = df_input.astype(int)
    I = df_input.values

    rc2 = RC2(WCNF())

    num_cells = len(cells)
    num_mutations = len(snvs)
    B_absent = np.zeros((num_cells, num_mutations), dtype=np.float64)
    B_present = np.zeros((num_cells, num_mutations), dtype=np.float64)

    Y = np.empty((num_cells, num_mutations), dtype=np.int64)
    numVarY = 0
    for i in range(num_cells):
        for j in range(num_mutations):
            numVarY += 1
            Y[i, j] = numVarY

    B = np.empty((num_mutations, num_mutations, 2, 2), dtype=np.int64)
    numVarB = 0
    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            for i, j in [(0, 1), (1, 0), (1, 1)]:
                numVarB += 1
                B[p, q, i, j] = numVarY + numVarB

    Z = np.empty((num_cells, num_mutations), dtype=np.int64)
    numVarZ = 0
    for i in range(num_cells):
        for j in range(num_mutations):
            if I[i, j] == 0:
                numVarZ += 1
                Z[i, j] = numVarY + numVarB + numVarZ

    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            rc2.add_clause([-B[p, q, 0, 1], -B[p, q, 1, 0], -B[p, q, 1, 1]])
            for i in range(num_cells):
                rc2.add_clause([-Y[i, p], -Y[i, q], B[p, q, 1, 1]])
                rc2.add_clause([Y[i, p], -Y[i, q], B[p, q, 0, 1]])
                rc2.add_clause([-Y[i, p], Y[i, q], B[p, q, 1, 0]])

    for i in range(num_cells):
        for j in range(num_mutations):
            ##0->1
            if alpha == 0:
                if I[i, j] == 0:
                    rc2.add_clause([-Y[i, j]], weight=1)
                if I[i, j] == 1:
                    rc2.add_clause([Y[i, j]])

            ##0->1 and 1->0
            if alpha > 0:
                if I[i, j] == 0:
                    rc2.add_clause([Y[i, j], Z[i, j]])
                    rc2.add_clause([-Y[i, j], -Z[i, j]])
                    rc2.add_clause([Z[i, j]], weight=math.log((1 - alpha) / beta))

                if I[i, j] == 1:
                    rc2.add_clause([Y[i, j]], weight=math.log((1 - beta) / alpha))

    s_time = time.time()
    variables = rc2.compute()
    e_time = time.time()
    running_time = e_time - s_time

    sol_Y = np.empty((num_cells, num_mutations), dtype=np.int8)
    numVar = 0
    for i in range(num_cells):
        for j in range(num_mutations):
            sol_Y[i, j] = variables[numVar] > 0
            numVar += 1

    df_output = pd.DataFrame(sol_Y)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    if not experiment:
        tsc.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output


def phiscsi(df_input, alpha, beta, time_out=86400):
    gp, gp_is_not_imported = tsc.ul.import_gurobi()
    if gp_is_not_imported:
        raise RuntimeError("Unable to import a package!")

    tsc.logg.info(
        f"running PhISCS-I with alpha={alpha}, beta={beta}, time_out={time_out}"
    )
    cells = list(df_input.index)
    snvs = list(df_input.columns)
    df_input = df_input.replace("?", 3)
    df_input = df_input.astype(int)
    I = df_input.values

    model = gp.Model("ILP")
    model.Params.OutputFlag = 0
    model.Params.LogFile = ""
    model.Params.Threads = 1
    model.Params.TimeLimit = time_out

    num_cells = len(cells)
    num_mutations = len(snvs)
    Y = {}
    B = {}
    for c in range(num_cells):
        for m in range(num_mutations):
            Y[c, m] = model.addVar(vtype=gp.GRB.BINARY, name="Y({0},{1})".format(c, m))
    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            B[p, q, 1, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name="B[{0},{1},1,1]".format(p, q)
            )
            B[p, q, 1, 0] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name="B[{0},{1},1,0]".format(p, q)
            )
            B[p, q, 0, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name="B[{0},{1},0,1]".format(p, q)
            )
    for p in range(num_mutations):
        for q in range(p + 1, num_mutations):
            model.addConstr(B[p, q, 0, 1] + B[p, q, 1, 0] + B[p, q, 1, 1] <= 2)
            for i in range(num_cells):
                model.addConstr(Y[i, p] + Y[i, q] - B[p, q, 1, 1] <= 1)
                model.addConstr(-Y[i, p] + Y[i, q] - B[p, q, 0, 1] <= 0)
                model.addConstr(Y[i, p] - Y[i, q] - B[p, q, 1, 0] <= 0)

    objective = 0
    for j in range(num_mutations):
        for i in range(num_cells):
            ##0->1 & 1->0
            if I[i, j] == 0:
                objective += np.log(1 - alpha) + np.log(beta / (1 - alpha)) * Y[i, j]
            if I[i, j] == 1:
                objective += np.log(alpha) + np.log((1 - beta) / alpha) * Y[i, j]

    model.setObjective(objective, gp.GRB.MAXIMIZE)

    s_time = time.time()
    model.optimize()
    e_time = time.time()
    running_time = e_time - s_time

    sol_Y = np.zeros((num_cells, num_mutations), dtype=np.int8)
    for i in range(num_cells):
        for j in range(num_mutations):
            sol_Y[i, j] = Y[i, j].X > 0.5

    df_output = pd.DataFrame(sol_Y)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    tsc.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output


def phiscs_original(
    df_input,
    alpha,
    beta,
    kmax=0,
    kel_weight=0,
    vaf_info=None,
    delta=0.2,
    time_out=86400,
):

    gp, gp_is_not_imported = tsc.ul.import_gurobi()
    if gp_is_not_imported:
        raise RuntimeError("Unable to import a package!")

    tsc.logg.info(
        f"running PhISCS-I with alpha={alpha}, beta={beta}, time_out={time_out}"
    )

    cells = list(df_input.index)
    snvs = list(df_input.columns)
    df_input = df_input.replace("?", 3)
    df_input = df_input.astype(int)
    I = df_input.values

    model = gp.Model("ILP")
    model.Params.OutputFlag = 0
    model.Params.LogFile = ""
    model.Params.Threads = 1
    model.Params.TimeLimit = time_out

    numCells = len(cells)
    numMutations = len(snvs)
    if vaf_info is not None:
        vaf_info = vaf_info.loc[snvs]
        sampleIDs = vaf_info.columns
        vaf_info.loc["NULL"] = vaf_info.shape[1] * [1]

    # --- Matrix Y is matrix of corrected (i.e. true) genotypes w.r.t. input SC matrix I
    Y = {}
    for c in range(numCells):
        for m in range(numMutations):
            Y[c, m] = model.addVar(vtype=gp.GRB.BINARY, name="Y({0},{1})".format(c, m))

    # --- Variables B control the existence of conflict between columns
    B = {}
    for p in range(numMutations + 1):
        for q in range(numMutations + 1):
            B[p, q, 1, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name="B[{0},{1},1,1]".format(p, q)
            )
            B[p, q, 1, 0] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name="B[{0},{1},1,0]".format(p, q)
            )
            B[p, q, 0, 1] = model.addVar(
                vtype=gp.GRB.BINARY, obj=0, name="B[{0},{1},0,1]".format(p, q)
            )

    K = {}
    for m in range(numMutations + 1):
        K[m] = model.addVar(vtype=gp.GRB.BINARY, name="K[{0}]".format(m))
    model.addConstr(K[numMutations] == 0)  # null mutation can not be eliminated

    # --- A[p,q] = 1 if p is ancestor of q
    A = {}
    if vaf_info is not None:
        for p in range(
            numMutations + 1
        ):  # mutation with index numMutation is null mutation
            for q in range(numMutations + 1):
                A[p, q] = model.addVar(
                    vtype=gp.GRB.BINARY, obj=0, name="A[{0},{1}]".format(p, q)
                )

    model.update()

    # --- number of eliminated columns is upper bounded by user provided constant
    model.addConstr(gp.quicksum(K[m] for m in range(numMutations)) <= kmax)

    # --- Enforce three gametes rule
    for i in range(numCells):
        for p in range(numMutations):
            for q in range(numMutations):
                model.addConstr(Y[i, p] + Y[i, q] - B[p, q, 1, 1] <= 1)
                model.addConstr(-Y[i, p] + Y[i, q] - B[p, q, 0, 1] <= 0)
                model.addConstr(Y[i, p] - Y[i, q] - B[p, q, 1, 0] <= 0)

    # --- Null mutation present in each cell
    for p in range(numMutations + 1):
        model.addConstr(B[p, numMutations, 1, 0] == 0)

    # --- Forbid conflict between columns (three gametes rule)
    for p in range(numMutations):
        for q in range(numMutations):
            model.addConstr(
                B[p, q, 0, 1] + B[p, q, 1, 0] + B[p, q, 1, 1] <= 2 + K[p] + K[q]
            )

    # --- Constraints for integrating VAF obtained from bulk data into the model
    if vaf_info is not None:
        for p in range(numMutations):
            for q in range(p + 1, numMutations):
                model.addConstr(A[p, q] + A[q, p] <= 1 - K[p])
                model.addConstr(A[p, q] + A[q, p] <= 1 - K[q])
                model.addConstr(A[p, q] + A[q, p] >= B[p, q, 1, 1] - K[p] - K[q])
        for p in range(numMutations + 1):
            model.addConstr(A[p, p] == 0)
            for q in range(numMutations + 1):
                model.addConstr(A[p, q] <= 1 - K[p])
                model.addConstr(A[p, q] <= 1 - K[q])

                if p < q:
                    model.addConstr(A[p, q] + A[q, p] <= 1)

                model.addConstr(A[p, q] + B[p, q, 0, 1] <= 1 + K[p] + K[q])
                model.addConstr(
                    B[p, q, 1, 0] + B[p, q, 1, 1] - A[p, q] <= 1 + K[p] + K[q]
                )

                for sampleID in sampleIDs:
                    VAF_p = float(vaf_info.iloc[p][sampleID])
                    VAF_q = float(vaf_info.iloc[q][sampleID])
                    model.addConstr(A[p, q] * VAF_p * (1 + delta) >= A[p, q] * VAF_q)

                    #'''
                    for r in range(numMutations + 1):
                        if r == q:
                            continue
                        VAF_r = float(vaf_info.iloc[r][sampleID])
                        # Constraint 2
                        model.addConstr(
                            VAF_p * (1 + delta)
                            >= VAF_q * (A[p, q] - A[r, q] - A[q, r])
                            + VAF_r * (A[p, r] - A[r, q] - A[q, r])
                        )
                    #'''

                for r in range(numMutations + 1):
                    if r == q:
                        continue
                    # Constraint 1.d
                    model.addConstr(A[p, r] >= A[p, q] + A[q, r] - 1)

            candidateAncestors = [i for i in range(numMutations + 1)]
            candidateAncestors.remove(p)

            if p < numMutations:
                model.addConstr(
                    gp.quicksum(A[s, p] for s in candidateAncestors) >= 1 - K[p]
                )
            elif p == numMutations:
                model.addConstr(gp.quicksum(A[s, p] for s in candidateAncestors) == 0)
            else:
                raise Exception("p index is out of range")

    # --- Defining the objective function
    objective = 0
    for j in range(numMutations):
        numZeros = 0
        numOnes = 0
        for i in range(numCells):
            if I[i][j] == 0:
                numZeros += 1
                objective += np.log(beta / (1 - alpha)) * Y[i, j]
            elif I[i][j] == 1:
                numOnes += 1
                objective += np.log((1 - beta) / alpha) * Y[i, j]

        objective += numZeros * np.log(1 - alpha)
        objective += numOnes * np.log(alpha)
        objective -= K[j] * (
            numZeros * np.log(1 - alpha)
            + numOnes * (np.log(alpha) + np.log((1 - beta) / alpha))
        )

    model.setObjective(objective, gp.GRB.MAXIMIZE)

    s_time = time.time()
    model.optimize()
    e_time = time.time()
    running_time = e_time - s_time

    removedMutsIDs = []
    sol_K = []
    for j in range(numMutations):
        sol_K.append(K[j].X > 0.5)
        if sol_K[j] == 1:
            removedMutsIDs.append(snvs[j])

    sol_Y = np.zeros((numCells, numMutations), dtype=np.int8)
    for i in range(numCells):
        for j in range(numMutations):
            sol_Y[i, j] = Y[i, j].X > 0.5

    df_output = pd.DataFrame(sol_Y)
    df_output.columns = snvs
    df_output.index = cells
    df_output.index.name = "cellIDxmutID"

    df_output[removedMutsIDs] = 0
    # df_output.drop(removedMutsIDs, axis=1, inplace=True)

    tsc.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



from collections import OrderedDict

import click

import trisicell as tsc
from trisicell.tl import consensus

from ._bnb import bnb
from ._booster import booster
from ._consensus import consensus
from ._huntress import huntress
from ._mcalling import mcalling
from ._others import cf2newick, cf2tree
from ._phiscs import phiscsb, phiscsi
from ._scistree import scistree
from ._scite import scite
from ._score import score


class NaturalOrderGroup(click.Group):
    """Command group trying to list subcommands in the order they were added.
    Make sure you initialize the `self.commands` with OrderedDict instance.
    With decorator, use::
        @click.group(cls=NaturalOrderGroup, commands=OrderedDict())
    """

    def list_commands(self, ctx):
        """List command names as they are in commands dict.
        If the dict is OrderedDict, it will preserve the order commands
        were added.
        """
        return self.commands.keys()


@click.version_option(version=tsc.__version__)
@click.group(
    cls=NaturalOrderGroup,
    commands=OrderedDict(),
    context_settings=dict(max_content_width=300, terminal_width=300),
)
def cli():
    """Scalable tumor phylogeny reconstruction and validation from single-cell data"""
    return None


cli.add_command(mcalling)
cli.add_command(score)
cli.add_command(scistree)
cli.add_command(scite)
cli.add_command(booster)
cli.add_command(phiscsb)
cli.add_command(phiscsi)
cli.add_command(bnb)
cli.add_command(huntress)
cli.add_command(cf2newick)
cli.add_command(cf2tree)
cli.add_command(consensus)



import os

import click

import trisicell as tsc


@click.command(short_help="Run PhISCS-BnB.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.option(
    "--bounding",
    "-b",
    default="real",
    type=click.Choice(["real", "simulated"]),
    show_default=True,
    help="Number of threads.",
)
def bnb(genotype_file, bounding):
    """PhISCS-BnB: a fast branch and bound algorithm for
    the perfect tumor phylogeny reconstruction problem :cite:`PhISCS-BnB`.

    trisicell bnb input.SC 0.0001 0.1 -b real
    """

    outfile = os.path.splitext(genotype_file)[0]

    tsc.settings.verbosity = "info"
    tsc.settings.logfile = f"{outfile}.bnb.log"

    df_in = tsc.io.read(genotype_file)
    df_out = tsc.tl.bnb(df_in, bounding=bounding)
    tsc.io.write(df_out, f"{outfile}.bnb.CFMatrix")

    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: Aug 14, 2020
# Description: mapping for the first pass of the STAR
# =========================================================================================

from trisicell.ul._servers import *


def run02(config, afterok):
    def cmds(sample):
        cmds = ""
        cmds += cmd([f"mkdir -p {config['outdir']}/{sample}"])
        if config["dotrimming"] == True:
            if config["pairedend"] == True:
                cmds += cmd(
                    [
                        f"trim_galore",
                        f"--gzip",
                        f"--length 30",
                        f"--fastqc",
                        f"--output_dir {config['outdir']}/{sample}",
                        f"--paired",
                        f"{config['infq']}/{config['infqpre1']}{sample}{config['infqpost1']} {config['infq']}/{config['infqpre2']}{sample}{config['infqpost2']}",
                    ]
                )
                if config["isrna"] == True:
                    cmds += cmd(
                        [
                            f"STAR",
                            f"--runMode alignReads",
                            f"--genomeDir {config['outdir']}/_indexing/1",
                            f"--sjdbGTFfile {config['annot']}",
                            f"--outFilterMultimapNmax 1",
                            f"--outSAMunmapped None",
                            f"--quantMode TranscriptomeSAM GeneCounts",
                            f"--runThreadN 1",
                            f"--sjdbOverhang {config['readlength']}",
                            f"--readFilesCommand zcat",
                            f"--outFileNamePrefix {config['outdir']}/{sample}/",
                            f"--readFilesIn {config['outdir']}/{sample}/{config['infqpre1']}{sample}{config['infqpost3']} {config['outdir']}/{sample}/{config['infqpre2']}{sample}{config['infqpost4']}",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"rm -rf",
                            f"{config['outdir']}/{sample}/Aligned.out.sam",
                        ]
                    )
                else:
                    cmds += cmd(
                        [
                            f"bwa mem",
                            f"-t 8",
                            f"{config['ref']}",
                            f"{config['outdir']}/{sample}/{config['infqpre1']}{sample}{config['infqpost3']} {config['outdir']}/{sample}/{config['infqpre2']}{sample}{config['infqpost4']}",
                            f"|",
                            f"samtools sort -@8 -m 10000000000",
                            f"-o {config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam -",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"samtools index",
                            f"{config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam",
                        ]
                    )
            else:
                cmds += cmd(
                    [
                        f"trim_galore",
                        f"--gzip",
                        f"--length 30",
                        f"--fastqc",
                        f"--output_dir {config['outdir']}/{sample}",
                        f"{config['infq']}/{config['infqpre']}{sample}{config['infqpost']}",
                    ]
                )
                if config["isrna"] == True:
                    cmds += cmd(
                        [
                            f"STAR",
                            f"--runMode alignReads",
                            f"--genomeDir {config['outdir']}/_indexing/1",
                            f"--sjdbGTFfile {config['annot']}",
                            f"--outFilterMultimapNmax 1",
                            f"--outSAMunmapped None",
                            f"--quantMode TranscriptomeSAM GeneCounts",
                            f"--runThreadN 1",
                            f"--sjdbOverhang {config['readlength']}",
                            f"--readFilesCommand zcat",
                            f"--outFileNamePrefix {config['outdir']}/{sample}/",
                            f"--readFilesIn {config['outdir']}/{sample}/{config['infqpre']}{sample}{config['infqpost5']}",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"rm -rf",
                            f"{config['outdir']}/{sample}/Aligned.out.sam",
                        ]
                    )
                else:
                    cmds += cmd(
                        [
                            f"bwa mem",
                            f"-t 8",
                            f"{config['ref']}",
                            f"{config['outdir']}/{sample}/{config['infqpre']}{sample}{config['infqpost5']}",
                            f"|",
                            f"samtools sort -@8 -m 10000000000",
                            f"-o {config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam -",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"samtools index",
                            f"{config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam",
                        ]
                    )
        else:
            if config["pairedend"] == True:
                if config["isrna"] == True:
                    cmds += cmd(
                        [
                            f"STAR",
                            f"--runMode alignReads",
                            f"--genomeDir {config['outdir']}/_indexing/1",
                            f"--sjdbGTFfile {config['annot']}",
                            f"--outFilterMultimapNmax 1",
                            f"--outSAMunmapped None",
                            f"--quantMode TranscriptomeSAM GeneCounts",
                            f"--runThreadN 1",
                            f"--sjdbOverhang {config['readlength']}",
                            f"--readFilesCommand zcat",
                            f"--outFileNamePrefix {config['outdir']}/{sample}/",
                            f"--readFilesIn {config['infq']}/{config['infqpre1']}{sample}{config['infqpost1']} {config['infq']}/{config['infqpre2']}{sample}{config['infqpost2']}",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"rm -rf",
                            f"{config['outdir']}/{sample}/Aligned.out.sam",
                        ]
                    )
                else:
                    cmds += cmd(
                        [
                            f"bwa mem",
                            f"-t 8",
                            f"{config['ref']}",
                            f"{config['infq']}/{config['infqpre1']}{sample}{config['infqpost1']} {config['infq']}/{config['infqpre2']}{sample}{config['infqpost2']}",
                            f"|",
                            f"samtools sort -@8 -m 10000000000",
                            f"-o {config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam -",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"samtools index",
                            f"{config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam",
                        ]
                    )
            else:
                if config["isrna"] == True:
                    cmds += cmd(
                        [
                            f"STAR",
                            f"--runMode alignReads",
                            f"--genomeDir {config['outdir']}/_indexing/1",
                            f"--sjdbGTFfile {config['annot']}",
                            f"--outFilterMultimapNmax 1",
                            f"--outSAMunmapped None",
                            f"--quantMode TranscriptomeSAM GeneCounts",
                            f"--runThreadN 1",
                            f"--sjdbOverhang {config['readlength']}",
                            f"--readFilesCommand zcat",
                            f"--outFileNamePrefix {config['outdir']}/{sample}/",
                            f"--readFilesIn {config['infq']}/{config['infqpre']}{sample}{config['infqpost']}",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"rm -rf",
                            f"{config['outdir']}/{sample}/Aligned.out.sam",
                        ]
                    )
                else:
                    cmds += cmd(
                        [
                            f"bwa mem",
                            f"-t 8",
                            f"{config['ref']}",
                            f"{config['infq']}/{config['infqpre']}{sample}{config['infqpost']}",
                            f"|",
                            f"samtools sort -@8 -m 10000000000",
                            f"-o {config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam -",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"samtools index",
                            f"{config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam",
                        ]
                    )
        cmds += cmd([f"echo Done!"], islast=True)
        return cmds

    df = pd.DataFrame()
    df["sample"] = config["samples"]
    df["cmd"] = df.apply(lambda x: cmds(x["sample"]), axis=1)

    jobname = os.path.basename(__file__)[:-3]
    if config["isrna"] == True:
        cmdmain = write_cmds_get_main(
            df,
            jobname,
            config["time02"],
            config["mem02"],
            "STAR/2.7.3a,trimmomatic/0.39,trimgalore/0.6.5,cutadapt/2.9,fastqc/0.11.9",
            1,
            config["email"],
            config["tmpdir"],
            afterok,
        )
    else:
        cmdmain = write_cmds_get_main(
            df,
            jobname,
            config["time02"],
            config["mem02"],
            "bwa/0.7.17,samtools/1.11,trimmomatic/0.39,trimgalore/0.6.5,cutadapt/2.9,fastqc/0.11.9",
            8,
            config["email"],
            config["tmpdir"],
            afterok,
        )
    return cmdmain



from trisicell.ul._dists import (
    calc_dist_by_cosine_ignore_na,
    calc_dist_by_dendro,
    calc_dist_by_l1_ignore_na,
)
from trisicell.ul._hclust import calc_hclust
from trisicell.ul._packages import *
from trisicell.ul._trees import to_cfmatrix, to_tree
from trisicell.ul._utils import *



import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import natsort as ns
import numpy as np
import pandas as pd
import seaborn as sns

import trisicell as tsc


def draw(adata, figsize=(10, 2), dpi=100):
    gspec = plt.GridSpec(1, 2, plt.figure(None, figsize, dpi=dpi))
    ax1 = plt.subplot(gspec[0])

    ax1 = plt.subplot(gspec[1])
    ax2.margins(0)

    return [ax1, ax2]


def plot_size(df_in):
    plt.rcParams["xtick.labelsize"] = 6
    plt.rcParams["ytick.labelsize"] = 6
    plt.figure(figsize=(1, 1))
    ax = sns.histplot([], color="black")
    ax.set_xlabel(f"{df_in.shape[1]} mutations", fontsize=8)
    ax.set_ylabel(f"{df_in.shape[0]} cells", fontsize=8)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.xaxis.set_label_position("top")


def plot_input(df_in):
    plt.rcParams["xtick.labelsize"] = 6
    plt.rcParams["ytick.labelsize"] = 6
    plt.figure(figsize=(1, 1))
    df = pd.DataFrame(np.array(np.unique(df_in.values, return_counts=True)).T)
    df[2] = 100 * df[1] / df_in.size
    df = df.set_index([0]).rename({0: "0", 1: "1", 3: "NA"})
    if "NA" in df.index:
        ax = df[2].plot.barh(color="#D92347", width=0.7)
    else:
        ax = df[2].plot.barh(color="#D92347", width=0.5)
    plt.gca().invert_yaxis()
    plt.xlim(0, 100)
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticks(np.arange(0, 101, 25))
    ax.tick_params(axis="x", direction="out", length=2, width=1)
    ax.tick_params(axis="y", direction="out", length=2, width=1)


def plot_output(df_out):
    plt.rcParams["xtick.labelsize"] = 6
    plt.rcParams["ytick.labelsize"] = 6
    plt.figure(figsize=(1, 1))
    df = pd.DataFrame(np.array(np.unique(df_out.values, return_counts=True)).T)
    df[2] = 100 * df[1] / df_out.size
    df = df.set_index([0]).rename({0: "0", 1: "1", 3: "NA"})
    ax = df[2].plot.barh(color="#0067A5", width=0.5)
    plt.gca().invert_yaxis()
    plt.xlim(0, 100)
    ax.set_xlabel("")
    ax.set_ylabel("")
    ax.set_xticks(np.arange(0, 101, 25))
    ax.tick_params(axis="x", direction="out", length=2, width=1)
    ax.tick_params(axis="y", direction="out", length=2, width=1)


def plot_flips2(df_in, df_out):
    I = df_in
    O = df_out
    df = pd.DataFrame(0, index=["0->1", "1->0", "NA->0", "NA->1"], columns=[1])
    df.loc["0->1", 1] = ((I == 0) & (O == 1)).sum().sum()
    df.loc["1->0", 1] = ((I == 1) & (O == 0)).sum().sum()
    df.loc["NA->0", 1] = ((I == 3) & (O == 0)).sum().sum()
    df.loc["NA->1", 1] = ((I == 3) & (O == 1)).sum().sum()
    print(df)


def plot_flips(df_in, df_out, row_colors):
    snvs = np.intersect1d(df_in.columns, df_out.columns)
    # cells = np.intersect1d(df_in.index, df_out.index)
    tmp = []
    for x in snvs:
        tmp.append(".".join(x.split(".chr")[1].split(".")[:2]))
    tmp = ns.index_natsorted(tmp)
    snvs = snvs[tmp]
    I = df_in.loc[row_colors.keys(), snvs]
    O = df_out.loc[row_colors.keys(), snvs]

    ccolors = []
    chrs = set()
    for x in snvs:
        c = x.split(".chr")[1].split(".")[0]
        if c in [f"{i}" for i in range(1, 22, 2)] + ["Y"]:
            ccolors.append("#969696")
            chrs.add(c)
        if c in [f"{i}" for i in range(2, 22, 2)] + ["X"]:
            ccolors.append("#252525")
            chrs.add(c)

    D = 1 * (I.values == 1) * (O.values == 0) + 2 * (I.values == 0) * (O.values == 1)
    rvb = mcolors.LinearSegmentedColormap.from_list(
        "", ["#DEDEDE", "#D92347", "#A9D0F5", "#FFFFFF"]
    )

    # x = I.shape[0] / min(I.shape[0], I.shape[1])
    # y = I.shape[1] / min(I.shape[0], I.shape[1])

    return sns.clustermap(
        D,
        vmin=0,
        vmax=3,
        metric="euclidean",
        cmap=rvb,
        row_cluster=False,
        col_cluster=False,
        row_colors=[row_colors.values()],
        col_colors=[ccolors],
        cbar_pos=None,
        figsize=(11, 2),
        xticklabels=False,
        yticklabels=False,
        # linecolor='white',
        # linewidths=0.8,
        # annot=False,
        colors_ratio=(0.01, 0.05),
        dendrogram_ratio=0,
    )


def plot_noisy(df_in, row_colors=None):
    if row_colors is None:
        row_colors = {x: "#000000" for x in df_in.index}
    snvs = df_in.columns
    tmp = []
    for x in snvs:
        tmp.append(".".join(x.split(".chr")[1].split(".")[:2]))
    tmp = ns.index_natsorted(tmp)
    snvs = snvs[tmp]

    ccolors = []
    chrs = set()
    for x in snvs:
        c = x.split(".chr")[1].split(".")[0]
        if c in [f"{i}" for i in range(1, 22, 2)] + ["Y"]:
            ccolors.append("#969696")
            chrs.add(c)
        if c in [f"{i}" for i in range(2, 22, 2)] + ["X"]:
            ccolors.append("#252525")
            chrs.add(c)

    D = df_in.loc[row_colors.keys(), snvs]
    D.index.name = ""
    rvb = mcolors.LinearSegmentedColormap.from_list(
        "", ["#DEDEDE", "#000000", "#FFFFFF"]
    )

    return sns.clustermap(
        D,
        vmin=0,
        vmax=3,
        metric="euclidean",
        cmap=rvb,
        row_cluster=False,
        col_cluster=False,
        row_colors=[row_colors.values()],
        col_colors=[ccolors],
        cbar_pos=None,
        figsize=(8, 2),
        xticklabels=False,
        yticklabels=False,
        # linecolor='white',
        # linewidths=0.8,
        # annot=False,
        colors_ratio=(0.01, 0.05),
        dendrogram_ratio=0,
    )



from pathlib import Path

from setuptools import find_packages, setup

try:
    from trisicell import __author__, __email__, __maintainer__, __version__
except ImportError:
    __author__ = ", ".join(["Farid Rashidi"])
    __maintainer__ = ", ".join(["Farid Rashidi"])
    __email__ = ", ".join(["farid.rsh@gmail.com"])
    __version__ = "0.0.3"

if __name__ == "__main__":
    setup(
        name="trisicell",
        entry_points="""
            [console_scripts]
            trisicell=trisicell.commands.trisicell:cli
        """,
        # ext_modules=extensions,
        # include_dirs=[np.get_include()],
        # package_data=package_data,
        use_scm_version=True,
        setup_requires=["setuptools_scm"],
        python_requires=">=3.6",
        install_requires=[
            l.strip() for l in Path("requirements.txt").read_text("utf-8").splitlines()
        ],
        dependency_links=["https://pypi.gurobi.com"],
        extras_require=dict(
            dev=[
                "black==20.8b1",
                "pre-commit==2.9.3",
                "isort>=5.7.0",
                "pytest-cov",
            ],
            docs=[
                l.strip()
                for l in (Path("docs") / "requirements.txt")
                .read_text("utf-8")
                .splitlines()
                if not l.startswith("-r")
            ],
        ),
        platforms=["Linux", "MacOSX"],
        packages=find_packages(),
        author=__author__,
        author_email=__email__,
        email=__email__,
        maintainer=__maintainer__,
        maintainer_email=__email__,
        version=__version__,
        description=Path("README.rst").read_text("utf-8").split("\n")[3],
        long_description=Path("README.rst").read_text("utf-8"),
        long_description_content_type="text/x-rst; charset=UTF-8",
        license="BSD",
        url="https://github.com/faridrashidi/trisicell",
        project_urls={
            "Documentation": "https://trisicell.readthedocs.io/en/latest",
            "Source Code": "https://github.com/faridrashidi/trisicell",
        },
        download_url="https://github.com/faridrashidi/trisicell",
        keywords=[
            "tumor phylogeny",
            "single cell",
            "scalable",
            "rna-seq",
            "dna-seq",
        ],
        classifiers=[
            "License :: OSI Approved :: BSD License",
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Natural Language :: English",
            "Framework :: Jupyter",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Scientific/Engineering :: Visualization",
        ],
    )



import click

import trisicell as tsc


@click.command(short_help="Calculate consensus betweeen two trees.")
@click.argument(
    "first_tree",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "second_tree",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "consensus_path",
    required=True,
    type=click.Path(exists=False),
)
def consensus(first_tree, second_tree, consensus_path):
    """The consensus tree between two phylogenetic trees.

    It writes the conflict-free matrix representing the consensus tree
    into the `consensus_path` filepath.

    trisicell consensus first_tree.CFMatrix second_tree.CFMatrix
    """

    tsc.settings.verbosity = "info"

    sc1 = tsc.io.read(first_tree)
    sc2 = tsc.io.read(second_tree)
    final_tree = tsc.tl.consensus_tree(sc1, sc2)
    tsc.io.write(final_tree.graph["data"], consensus_path)

    return None



import os

import click

import trisicell as tsc


@click.command(short_help="Run PhISCS (CSP version).")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
def phiscsb(genotype_file, alpha, beta):
    """PhISCS: a combinatorial approach for subperfect
    tumor phylogeny reconstructionvia integrative use of
    single-cell and bulk sequencing data :cite:`PhISCS`.

    trisicell phiscsb input.SC 0.0001 0.1
    """

    outfile = os.path.splitext(genotype_file)[0]

    tsc.settings.verbosity = "info"
    tsc.settings.logfile = f"{outfile}.phiscsb.log"

    df_in = tsc.io.read(genotype_file)
    df_out = tsc.tl.phiscsb(df_in, alpha=alpha, beta=beta)
    tsc.io.write(df_out, f"{outfile}.phiscsb.CFMatrix")

    return None


@click.command(short_help="Run PhISCS (ILP version).")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--time_out",
    "-to",
    default=86400,
    type=int,
    show_default=True,
    help="Timelimit of the program.",
)
def phiscsi(genotype_file, alpha, beta, time_out):
    """PhISCS: a combinatorial approach for subperfect
    tumor phylogeny reconstructionvia integrative use of
    single-cell and bulk sequencing data :cite:`PhISCS`.

    trisicell phiscsi input.SC 0.0001 0.1 -to 3600
    """

    outfile = os.path.splitext(genotype_file)[0]

    tsc.settings.verbosity = "info"
    tsc.settings.logfile = f"{outfile}.phiscsi.log"

    df_in = tsc.io.read(genotype_file)
    df_out = tsc.tl.phiscsi(df_in, alpha=alpha, beta=beta, time_out=time_out)
    tsc.io.write(df_out, f"{outfile}.phiscsi.CFMatrix")

    return None



import copy
import itertools
import time

import numpy as np
import pandas as pd
import pybnb
import scipy.sparse as sp
from pysat.examples.rc2 import RC2
from pysat.formula import WCNF

import trisicell as tsc

rec_num = 0


def bnb(df_input, bounding):
    mpi4py, mpi4py_is_not_imporeted = tsc.ul.import_mpi4py()
    if mpi4py_is_not_imporeted:
        raise RuntimeError("Unable to import a package!")

    tsc.logg.info(f"running BnB with bounding={bounding}")
    matrix_input = df_input.values
    matrix_output = matrix_input.copy()
    na_value = 3
    bounding_algs = {
        "real": TwoSatBounding(
            heuristic_setting=None,
            n_levels=2,
            compact_formulation=False,
            na_value=na_value,
        ),  # Real Data
        "simulated": TwoSatBounding(
            heuristic_setting=[True, True, False, True, True],
            n_levels=1,
            compact_formulation=True,
            na_value=na_value,
        ),  # Simulation
    }

    bounding_alg = bounding_algs[bounding]

    s_time = time.time()
    flips = solve_by_BnB(matrix_input, na_value, bounding_alg)
    e_time = time.time()
    running_time = e_time - s_time

    for k in flips:
        matrix_output[k] = 1
    matrix_output[np.where(matrix_output == na_value)] = 0

    df_output = pd.DataFrame(matrix_output)
    df_output.columns = df_input.columns
    df_output.index = df_input.index
    df_output.index.name = "cellIDxmutID"

    tsc.ul.stat(df_input, df_output, 0.1, 0.000000001, running_time)

    return df_output


def solve_by_BnB(matrix_in, na_value, bounding_alg):
    result = bnb_solve(matrix_in, bounding_algorithm=bounding_alg, na_value=na_value)
    matrix_output = result[0]
    flips = []
    zero_one_flips = np.where((matrix_in != matrix_output) & (matrix_in != na_value))
    for i in range(len(zero_one_flips[0])):
        flips.append((zero_one_flips[0][i], zero_one_flips[1][i]))
    na_one_flips = np.where((matrix_output == 1) & (matrix_in == na_value))
    for i in range(len(na_one_flips[0])):
        flips.append((na_one_flips[0][i], na_one_flips[1][i]))
    return flips


def all_None(*args):
    return args.count(None) == len(args)


def calculate_column_intersections(matrix, for_loop=False, row_by_row=False):
    ret = np.empty((matrix.shape[1], matrix.shape[1]), dtype=np.bool)
    mask_1 = matrix == 1

    if for_loop:
        for p in range(matrix.shape[1]):
            # even though the diagonals are not necessary, I keep it for ease of debugging
            for q in range(p, matrix.shape[1]):
                ret[p, q] = np.any(np.logical_and(mask_1[:, p], mask_1[:, q]))
                ret[q, p] = ret[p, q]
    elif row_by_row:
        ret[:, :] = 0
        for r in range(matrix.shape[0]):
            one_columns = mask_1[r]
            ret[np.ix_(one_columns, one_columns)] = True
    return ret


def zero_or_na(vec, na_value=-1):
    return np.logical_or(vec == 0, vec == na_value)


def make_sure_variable_exists(
    memory_matrix, row, col, num_var_F, map_f2ij, var_list, na_value
):
    if memory_matrix[row, col] < 0:
        num_var_F += 1
        map_f2ij[num_var_F] = (row, col)
        memory_matrix[row, col] = num_var_F
        var_list.append(num_var_F)
    return num_var_F


def get_effective_matrix(I, delta01, delta_na_to_1, change_na_to_0=False):
    x = np.array(I + delta01, dtype=np.int8)
    if delta_na_to_1 is not None:
        na_indices = delta_na_to_1.nonzero()
        x[
            na_indices
        ] = 1  # should have been (but does not accept): x[na_indices] = delta_na_to_1[na_indices]
    if change_na_to_0:
        x[np.logical_and(x != 0, x != 1)] = 0
    return x


def make_twosat_model_from_np(
    constraints,
    F,
    zero_vars,
    na_vars,
    eps=None,
    heuristic_setting=None,
    compact_formulation=True,
):

    if eps is None:
        eps = 1 / (len(zero_vars) + len(na_vars))

    if heuristic_setting is None:
        rc2 = RC2(WCNF())
    else:
        assert len(heuristic_setting) == 5
        rc2 = RC2(
            WCNF(),
            adapt=heuristic_setting[0],
            exhaust=heuristic_setting[1],
            incr=heuristic_setting[2],
            minz=heuristic_setting[3],
            trim=heuristic_setting[4],
        )

    if not compact_formulation:
        # hard constraints Z_a,p or Z_b,q
        for constr_ind in range(constraints[0].shape[0]):
            constraint = constraints[0][constr_ind]
            a, p, b, q = constraint.flat
            # print(constraint, F.shape)
            # print(a, p, b, q)
            rc2.add_clause([F[a, p], F[b, q]])
        if len(constraints) >= 2:
            # hard constraints Z_a,p or Z_b,q or -Z_c,d
            for constr_ind in range(constraints[1].shape[0]):
                constraint = constraints[1][constr_ind]
                a, p, b, q, c, d = constraint.flat
                # print(a, p, b, q, c, d)
                rc2.add_clause([F[a, p], F[b, q], -F[c, d]])
    else:
        # hard constraints Z_a,p or (sign) b_pq
        for constr_ind in range(constraints[0].shape[0]):
            constraint = constraints[0][constr_ind]
            row, col, b_pq, sign = constraint.flat
            rc2.add_clause([F[row, col], sign * b_pq])
        if len(constraints) >= 2:
            # hard constraints Z_a,p or Z_b,q or -Z_c,d
            for constr_ind in range(constraints[1].shape[0]):
                constraint = constraints[1][constr_ind]
                row, col, c_pq0, c_pq1 = constraint.flat
                # if Z_rc is True at least one of p, q should become active
                # E.g., c_pq0 be False
                rc2.add_clause([-F[row, col], -c_pq0, -c_pq1])
                # if c_pq0 is False then Z_rc has to be flipped
                rc2.add_clause([F[row, col], c_pq0])

    # soft constraints for zero variables
    for var in zero_vars:
        rc2.add_clause([-var], weight=1)

    if eps > 0:
        # soft constraints for zero variables
        for var in na_vars:
            rc2.add_clause([-var], weight=eps)

    return rc2


def twosat_solver(
    matrix,
    cluster_rows=False,
    cluster_cols=False,
    only_descendant_rows=False,
    na_value=None,
    leave_nas_if_zero=False,
    return_lb=False,
    heuristic_setting=None,
    n_levels=2,
    eps=0,
    compact_formulation=False,
):
    global rec_num
    rec_num += 1
    assert not cluster_rows, "Not implemented yet"
    assert not cluster_cols, "Not implemented yet"
    assert not only_descendant_rows, "Not implemented yet"
    model_time = 0
    opt_time = 0
    start_time = time.time()

    return_value = make_constraints_np_matrix(
        matrix,
        n_levels=n_levels,
        na_value=na_value,
        compact_formulation=compact_formulation,
    )
    model_time += time.time() - start_time
    F, map_f2ij, zero_vars, na_vars, hard_constraints, col_pair = (
        return_value.F,
        return_value.map_f2ij,
        return_value.zero_vars,
        return_value.na_vars,
        return_value.hard_constraints,
        return_value.col_pair,
    )

    if col_pair is not None:
        icf = False
    elif return_value.complete_version:
        icf = True
    else:
        icf = None

    final_output = None
    lower_bound = 0
    if icf:
        final_output, total_time = matrix.copy(), 0
    else:
        start_time = time.time()
        rc2 = make_twosat_model_from_np(
            hard_constraints,
            F,
            zero_vars,
            na_vars,
            eps,
            heuristic_setting,
            compact_formulation=compact_formulation,
        )
        model_time += time.time() - start_time

        a = time.time()
        variables = rc2.compute()
        b = time.time()
        opt_time += b - a
        output_matrix = matrix.copy()
        output_matrix = output_matrix.astype(np.int8)

        for var_ind in range(len(variables)):
            if (
                0 < variables[var_ind] and variables[var_ind] in map_f2ij
            ):  # if 0 or 2 make it one
                output_matrix[map_f2ij[variables[var_ind]]] = 1
                if matrix[map_f2ij[variables[var_ind]]] != na_value:
                    lower_bound += 1
        # I don't change 2s to 0s here keep them 2 for next time

        # For recursion I set off all sparsification parameters
        # Also I want na->0 to stay na for the recursion regardless of original input for leave_nas_if_zero
        # I am also not passing eps here to wrap up the recursion soon

        Orec, rec_model_time, rec_opt_time = twosat_solver(
            output_matrix,
            na_value=na_value,
            heuristic_setting=None,
            n_levels=n_levels,
            leave_nas_if_zero=True,
            compact_formulation=compact_formulation,
        )
        model_time += rec_model_time
        opt_time += rec_opt_time

        if not leave_nas_if_zero:
            Orec[Orec == na_value] = 0
        final_output = Orec

    if return_lb:
        return final_output, model_time, opt_time, lower_bound
    else:
        return final_output, model_time, opt_time


def make_constraints_np_matrix(
    matrix,
    constraints=None,
    n_levels=2,
    na_value=None,
    row_coloring=None,
    col_coloring=None,
    probability_threshold=None,
    fn_rate=None,
    column_intersection=None,
    compact_formulation=True,
):
    """
    Returns a "C x 2 x 2" matrix where C is the number of extracted constraints each constraints is of the form:
    ((r1, c1), (r2, c2)) and correspond to Z_{r1, c1} or Z{r2, c2}
    :param matrix: A binary matrix cellsXmutations
    :param constraints: If not None instead of evaluating the whole matrix it will only look at potential constraints
    :param level: The type of constraints to add
    :param na_value:
    :param row_coloring: Only constraints that has the same row coloring will be used
    :param col_coloring: Only constraints that has the same column coloring will be used
    :param probability_threshold:
    :param fn_rate:
    :return:
    """
    # TD: Take decendence analysis out of here?
    # TD: how to reuse constraints input
    from collections import namedtuple

    assert (probability_threshold is None) == (fn_rate is None)
    descendance_analysis = probability_threshold is not None
    assert 1 <= n_levels <= 2, "not implemented yet"

    # means none of scarification ideas have been used
    complete_version = all_None(
        row_coloring, col_coloring, probability_threshold, fn_rate
    )

    soft_cnst_num = 0
    hard_constraints = [[] for _ in range(n_levels)]  # an empty list each level
    if descendance_analysis:
        # dictionary for lazy calculation of decadence:
        descendent_dict = dict()

    # variables for each zero
    F = -np.ones(matrix.shape, dtype=np.int64)
    num_var_F = 0
    map_f2ij = dict()
    zero_vars = list()
    na_vars = list()
    if compact_formulation:
        B_vars_offset = matrix.shape[0] * matrix.shape[1] + 1
        num_var_B = 0
        map_b2ij = dict()
        if n_levels >= 2:
            C_vars_offset = B_vars_offset + matrix.shape[1] * matrix.shape[1] + 1
            num_var_C = 0
            map_c2ij = dict()

    col_pair = None
    pair_cost = 0

    if column_intersection is None:
        column_intersection = calculate_column_intersections(matrix, row_by_row=True)
        # column_intersection = calculate_column_intersections(matrix, for_loop=True)
    for p in range(matrix.shape[1]):
        for q in range(p + 1, matrix.shape[1]):
            if column_intersection[p, q]:  # p and q has intersection
                # TD: check col_coloring here
                r01 = np.nonzero(
                    np.logical_and(
                        zero_or_na(matrix[:, p], na_value=na_value), matrix[:, q] == 1
                    )
                )[0]
                r10 = np.nonzero(
                    np.logical_and(
                        matrix[:, p] == 1, zero_or_na(matrix[:, q], na_value=na_value)
                    )
                )[0]
                cost = min(len(r01), len(r10))
                if cost > pair_cost:  # keep best pair to return as auxiliary info
                    # print("------------", cost, (p, q), len(r01), len(r10), column_intersection[p, q])
                    col_pair = (p, q)
                    pair_cost = cost
                if cost > 0:  # don't do anything if one of r01 or r10 is empty
                    if (
                        not compact_formulation
                    ):  # len(r01) * len(r10) many constraints will be added
                        for a, b in itertools.product(r01, r10):
                            # TD: check row_coloring
                            for row, col in [
                                (a, p),
                                (b, q),
                            ]:  # make sure the variables for this are made
                                var_list = (
                                    zero_vars if matrix[row, col] == 0 else na_vars
                                )
                                num_var_F = make_sure_variable_exists(
                                    F, row, col, num_var_F, map_f2ij, var_list, na_value
                                )
                            hard_constraints[0].append(
                                [[a, p], [b, q]]
                            )  # at least one of them should be flipped
                    else:  # compact formulation: (r01 + r10) number of new constraints will be added
                        # define new B variable
                        b_pq = B_vars_offset + num_var_B
                        num_var_B += 1
                        for row_list, col, sign in zip((r01, r10), (p, q), (1, -1)):
                            for row in row_list:
                                var_list = (
                                    zero_vars if matrix[row, col] == 0 else na_vars
                                )
                                num_var_F = make_sure_variable_exists(
                                    F, row, col, num_var_F, map_f2ij, var_list, na_value
                                )
                                hard_constraints[0].append([row, col, b_pq, sign])
                                # this will be translated to (Z_ap or (sign)B_pq)
            elif n_levels >= 2:
                r01 = np.nonzero(
                    np.logical_and(
                        zero_or_na(matrix[:, p], na_value=na_value), matrix[:, q] == 1
                    )
                )[0]
                r10 = np.nonzero(
                    np.logical_and(
                        matrix[:, p] == 1, zero_or_na(matrix[:, q], na_value=na_value)
                    )
                )[0]
                cost = min(len(r01), len(r10))
                if cost > 0:  # don't do anything if one of r01 or r10 is empty
                    if not compact_formulation:
                        # len(r01) * len(r10) * (len(r01) * len(r10)) many constraints will be added
                        x = np.empty((r01.shape[0] + r10.shape[0], 2), dtype=np.int)
                        x[: len(r01), 0] = r01
                        x[: len(r01), 1] = p
                        x[-len(r10) :, 0] = r10
                        x[-len(r10) :, 1] = q

                        for a, b, ind in itertools.product(r01, r10, range(x.shape[0])):
                            for row, col in [
                                (a, p),
                                (b, q),
                                (x[ind, 0], x[ind, 1]),
                            ]:  # make sure the variables for this are made
                                # print(row, col)
                                var_list = (
                                    zero_vars if matrix[row, col] == 0 else na_vars
                                )
                                num_var_F = make_sure_variable_exists(
                                    F, row, col, num_var_F, map_f2ij, var_list, na_value
                                )
                            row = [[a, p], [b, q], [x[ind, 0], x[ind, 1]]]
                            if not np.array_equal(
                                row[0], row[2]
                            ) and not np.array_equal(row[1], row[2]):
                                hard_constraints[1].append(
                                    [[a, p], [b, q], [x[ind, 0], x[ind, 1]]]
                                )
                    else:  #  if compact_formulation: 2(r01 + r10) will be added
                        # define two new C variable
                        c_pq0 = C_vars_offset + num_var_C
                        num_var_C += 1
                        c_pq1 = C_vars_offset + num_var_C
                        num_var_C += 1
                        for row_list, col, sign in zip((r01, r10), (p, q), (1, -1)):
                            for row in row_list:
                                var_list = (
                                    zero_vars if matrix[row, col] == 0 else na_vars
                                )
                                num_var_F = make_sure_variable_exists(
                                    F, row, col, num_var_F, map_f2ij, var_list, na_value
                                )
                                if sign == 1:
                                    hard_constraints[1].append([row, col, c_pq0, c_pq1])
                                    # this will be translated to (~Z_ap or ~c_pq0 or ~c_pq1)
                                    # and (Z_ap or c_pq0)
                                else:
                                    hard_constraints[1].append([row, col, c_pq1, c_pq0])
                                    # this will be translated to (~Z_ap or ~c_pq0 or ~c_pq1) (the same)
                                    # and (Z_ap or c_pq1) (different)

    # TD: when using this make sure to put an if to say if the model is small and
    return_type = namedtuple(
        "ReturnType",
        "F map_f2ij zero_vars na_vars hard_constraints col_pair complete_version",
    )
    for ind in range(n_levels):
        hard_constraints[ind] = np.array(hard_constraints[ind], dtype=np.int)
    return return_type(
        F, map_f2ij, zero_vars, na_vars, hard_constraints, col_pair, complete_version
    )


def is_conflict_free_gusfield_and_get_two_columns_in_coflicts(I, na_value):
    def sort_bin(a):
        b = np.transpose(a)
        b_view = np.ascontiguousarray(b).view(
            np.dtype((np.void, b.dtype.itemsize * b.shape[1]))
        )
        idx = np.argsort(b_view.ravel())[::-1]
        c = b[idx]
        return np.transpose(c), idx

    Ip = I.copy()
    Ip[Ip == na_value] = 0
    O, idx = sort_bin(Ip)
    # TD: delete duplicate columns
    # print(O, '\n')
    Lij = np.zeros(O.shape, dtype=int)
    for i in range(O.shape[0]):
        maxK = 0
        for j in range(O.shape[1]):
            if O[i, j] == 1:
                Lij[i, j] = maxK
                maxK = j + 1
    # print(Lij, '\n')
    Lj = np.amax(Lij, axis=0)
    # print(Lj, '\n')
    for i in range(O.shape[0]):
        for j in range(O.shape[1]):
            if O[i, j] == 1:
                if Lij[i, j] != Lj[j]:
                    return False, (idx[j], idx[Lj[j] - 1])
    return True, (None, None)


class BoundingAlgAbstract:
    def __init__(self):
        self.matrix = None
        self._extra_info = None
        self._extraInfo = {}
        self._times = {}
        self.na_support = False
        pass

    def reset(self, matrix):
        raise NotImplementedError("The method not implemented")

    def get_bound(self, delta):
        """
        This bound should include the flips done so far too
        delta: a sparse matrix with fliped ones
        """
        raise NotImplementedError("The method not implemented")

    def get_name(self):
        return type(self).__name__

    def get_state(self):
        return None

    def set_state(self, state):
        assert state is None
        pass

    def get_extra_info(self):
        """
        Some bounding algorithms can provide extra information after calling bounding.
        E.g.,
        return {"icf":True, "bestPair":(a,b)}
        """
        return copy.copy(self._extraInfo)

    def get_priority(self, till_here, this_step, after_here, icf=False):
        return -after_here

    def get_times(self):
        return self._times

    def get_init_node(self):
        return None


class TwoSatBounding(BoundingAlgAbstract):
    def __init__(
        self,
        priority_version=-1,
        cluster_rows=False,
        cluster_cols=False,
        only_descendant_rows=False,
        na_value=None,
        heuristic_setting=None,
        n_levels=2,
        eps=0,
        compact_formulation=False,
    ):
        """
        :param priority_version:
        """
        assert not cluster_rows, "Not implemented yet"
        assert not cluster_cols, "Not implemented yet"
        assert not only_descendant_rows, "Not implemented yet"

        self.priority_version = priority_version

        self.na_support = True
        self.na_value = na_value
        self.matrix = None
        self._times = None
        self.next_lb = None
        self.heuristic_setting = heuristic_setting
        self.n_levels = n_levels
        self.eps = eps  # only for upperbound
        self.compact_formulation = compact_formulation
        self.cluster_rows = cluster_rows
        self.cluster_cols = cluster_cols
        self.only_descendant_rows = only_descendant_rows

    def get_name(self):
        params = [
            type(self).__name__,
            self.priority_version,
            self.heuristic_setting,
            self.n_levels,
            self.eps,
            self.compact_formulation,
        ]
        params_str = map(str, params)
        return "_".join(params_str)

    def reset(self, matrix):
        self.matrix = matrix  # TD: make the model here and do small alterations later

        # self.na_value = infer_na_value(matrix)
        self._times = {"model_preparation_time": 0, "optimization_time": 0}

    def get_init_node(self):

        # def twosat_solver(matrix, cluster_rows=False, cluster_cols=False, only_descendant_rows=False,
        #                   na_value=None, leave_nas_if_zero=False, return_lb=False, heuristic_setting=None,
        #                   n_levels=2, eps=0, compact_formulation=True):
        #     pass

        node = pybnb.Node()
        solution, model_time, opt_time, lb = twosat_solver(
            self.matrix,
            cluster_rows=self.cluster_rows,
            cluster_cols=self.cluster_cols,
            only_descendant_rows=self.only_descendant_rows,
            na_value=self.na_value,
            leave_nas_if_zero=True,
            return_lb=True,
            heuristic_setting=None,
            n_levels=self.n_levels,
            eps=self.eps,
            compact_formulation=self.compact_formulation,
        )
        self._times["model_preparation_time"] += model_time
        self._times["optimization_time"] += opt_time

        nodedelta = sp.lil_matrix(np.logical_and(solution == 1, self.matrix == 0))
        node_na_delta = sp.lil_matrix(
            np.logical_and(solution == 1, self.matrix == self.na_value)
        )
        node.state = (
            nodedelta,
            True,
            None,
            nodedelta.count_nonzero(),
            self.get_state(),
            node_na_delta,
        )
        node.queue_priority = self.get_priority(
            till_here=-1, this_step=-1, after_here=-1, icf=True
        )
        self.next_lb = lb
        return node

    def get_bound(self, delta, delta_na=None):
        # make this dynamic when more nodes were getting explored
        if self.next_lb is not None:
            lb = self.next_lb
            self.next_lb = None
            return lb
        self._extraInfo = None
        current_matrix = get_effective_matrix(self.matrix, delta, delta_na)
        has_na = np.any(current_matrix == self.na_value)

        model_time = time.time()
        return_value = make_constraints_np_matrix(
            current_matrix,
            n_levels=self.n_levels,
            na_value=self.na_value,
            compact_formulation=self.compact_formulation,
        )
        F, map_f2ij, zero_vars, na_vars, hard_constraints, col_pair = (
            return_value.F,
            return_value.map_f2ij,
            return_value.zero_vars,
            return_value.na_vars,
            return_value.hard_constraints,
            return_value.col_pair,
        )

        if col_pair is not None:
            icf = False
        elif return_value.complete_version:
            icf = True
        else:
            icf = None  # not sure
        rc2 = make_twosat_model_from_np(
            hard_constraints,
            F,
            zero_vars,
            na_vars,
            eps=0,
            heuristic_setting=self.heuristic_setting,
            compact_formulation=self.compact_formulation,
        )

        model_time = time.time() - model_time
        self._times["model_preparation_time"] += model_time

        opt_time = time.time()
        variables = rc2.compute()
        opt_time = time.time() - opt_time
        self._times["optimization_time"] += opt_time

        result = 0
        for var_ind in range(len(variables)):
            if (
                variables[var_ind] > 0
                and abs(variables[var_ind]) in map_f2ij
                and self.matrix[map_f2ij[abs(variables[var_ind])]] == 0
            ):
                result += 1

        assert has_na or ((result == 0) == (col_pair is None)), f"{result}_{col_pair}"
        self._extraInfo = {
            "icf": icf,
            "one_pair_of_columns": col_pair,
        }
        ret = result + delta.count_nonzero()
        return ret

    def get_priority(self, till_here, this_step, after_here, icf=False):
        if icf:
            return self.matrix.shape[0] * self.matrix.shape[1] + 10
        else:
            sgn = np.sign(self.priority_version)
            pv_abs = self.priority_version * sgn
            if pv_abs == 1:
                return sgn * (till_here + this_step + after_here)
            elif pv_abs == 2:
                return sgn * (this_step + after_here)
            elif pv_abs == 3:
                return sgn * (after_here)
            elif pv_abs == 4:
                return sgn * (till_here + after_here)
            elif pv_abs == 5:
                return sgn * (till_here)
            elif pv_abs == 6:
                return sgn * (till_here + this_step)
            elif pv_abs == 7:
                return 0
        assert False, "get_priority did not return anything!"


class BnB(pybnb.Problem):
    def __init__(self, I, boundingAlg: BoundingAlgAbstract, na_value=None):
        self.na_value = na_value
        self.has_na = np.any(I == self.na_value)
        self.I = I
        self.delta = sp.lil_matrix(I.shape, dtype=np.int8)  # this can be coo_matrix too
        self.boundingAlg = boundingAlg
        self.delta_na = None
        if self.has_na:
            assert (
                boundingAlg.na_support
            ), "Input has N/A coordinates but bounding algorithm doesn't support it."
            self.delta_na = sp.lil_matrix(
                I.shape, dtype=np.int8
            )  # the coordinates with na that are decided to be 1
        (
            self.icf,
            self.colPair,
        ) = is_conflict_free_gusfield_and_get_two_columns_in_coflicts(self.I, na_value)
        self.boundingAlg.reset(I)
        self.node_to_add = self.boundingAlg.get_init_node()
        self.bound_value = self.boundingAlg.get_bound(self.delta)

    def sense(self):
        return pybnb.minimize

    def objective(self):
        if self.icf:
            return self.delta.count_nonzero()
        else:
            return pybnb.Problem.infeasible_objective(self)

    def bound(self):
        return self.bound_value

    def save_state(self, node):
        node.state = (
            self.delta,
            self.icf,
            self.colPair,
            self.bound_value,
            self.boundingAlg.get_state(),
            self.delta_na,
        )

    def load_state(self, node):
        (
            self.delta,
            self.icf,
            self.colPair,
            self.bound_value,
            boundingAlgState,
            self.delta_na,
        ) = node.state
        self.boundingAlg.set_state(boundingAlgState)

    def get_current_matrix(self):
        return get_effective_matrix(self.I, self.delta, self.delta_na)

    def branch(self):
        if self.icf:
            return

        need_for_new_nodes = True
        if self.node_to_add is not None:
            newnode = self.node_to_add
            self.node_to_add = None
            if (
                newnode.state[0].count_nonzero() == self.bound_value
            ):  # current_obj == lb => no need to explore
                need_for_new_nodes = False
            assert (
                newnode.queue_priority is not None
            ), "Right before adding a node its priority in the queue is not set!"
            yield newnode

        if need_for_new_nodes:
            p, q = self.colPair
            nf01 = None
            current_matrix = self.get_current_matrix()
            for col, colp in [(q, p), (p, q)]:
                node = pybnb.Node()
                nodedelta = copy.deepcopy(self.delta)
                node_na_delta = copy.deepcopy(self.delta_na)
                col1 = np.array(current_matrix[:, col], dtype=np.int8).reshape(-1)
                col2 = np.array(current_matrix[:, colp], dtype=np.int8).reshape(-1)
                rows01 = np.nonzero(np.logical_and(col1 == 0, col2 == 1))[0]
                rows21 = np.nonzero(np.logical_and(col1 == self.na_value, col2 == 1))[0]
                if (
                    len(rows01) + len(rows21) == 0
                ):  # nothing has changed! Dont add new node
                    continue
                nodedelta[rows01, col] = 1
                nf01 = nodedelta.count_nonzero()
                if self.has_na:
                    node_na_delta[rows21, col] = 1
                    new_bound = self.boundingAlg.get_bound(nodedelta, node_na_delta)
                else:
                    new_bound = self.boundingAlg.get_bound(nodedelta)

                node_icf, nodecol_pair = None, None
                extra_info = self.boundingAlg.get_extra_info()

                if extra_info is not None:
                    if "icf" in extra_info:
                        node_icf = extra_info["icf"]
                    if "one_pair_of_columns" in extra_info:
                        nodecol_pair = extra_info["one_pair_of_columns"]
                if node_icf is None:
                    x = get_effective_matrix(self.I, nodedelta, node_na_delta)
                    (
                        node_icf,
                        nodecol_pair,
                    ) = is_conflict_free_gusfield_and_get_two_columns_in_coflicts(
                        x, self.na_value
                    )

                node_bound_value = max(self.bound_value, new_bound)
                node.state = (
                    nodedelta,
                    node_icf,
                    nodecol_pair,
                    node_bound_value,
                    self.boundingAlg.get_state(),
                    node_na_delta,
                )
                node.queue_priority = self.boundingAlg.get_priority(
                    till_here=nf01 - len(rows01),
                    this_step=len(rows01),
                    after_here=new_bound - nf01,
                    icf=node_icf,
                )
                assert (
                    node.queue_priority is not None
                ), "Right before adding a node its priority in the queue is not set!"
                yield node


def bnb_solve(matrix, bounding_algorithm, na_value=None):
    problem1 = BnB(matrix, bounding_algorithm, na_value=na_value)
    solver = pybnb.solver.Solver()
    results1 = solver.solve(problem1, queue_strategy="custom", log=None)
    if results1.solution_status != "unknown":
        returned_delta = results1.best_node.state[0]
        returned_delta_na = results1.best_node.state[-1]
        returned_matrix = get_effective_matrix(
            matrix, returned_delta, returned_delta_na, change_na_to_0=True
        )
    else:
        returned_matrix = np.zeros((1, 1))
    # print("results1.nodes:  ", results1.nodes)
    return returned_matrix, results1.termination_condition



import os

import click

import trisicell as tsc


@click.command(short_help="Run ScisTree.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
def scistree(genotype_file, alpha, beta):
    """Accurate and efficient cell lineage tree inference from noisy
    single cell data: the maximum likelihood perfect phylogeny approach :cite:`ScisTree`.

    trisicell scistree input.SC 0.0001 0.1
    """

    outfile = os.path.splitext(genotype_file)[0]

    tsc.settings.verbosity = "info"
    tsc.settings.logfile = f"{outfile}.scistree.log"

    df_in = tsc.io.read(genotype_file)
    df_out = tsc.tl.scistree(df_in, alpha=alpha, beta=beta, save_inter=False)
    tsc.io.write(df_out, f"{outfile}.scistree.CFMatrix")

    return None



import click

import trisicell as tsc


@click.command(short_help="Calculate scores between two trees.")
@click.argument(
    "ground_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "inferred_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
def score(ground_file, inferred_file):
    """Scores of Anscenstor-descendant, Different-lineage and MLTD.

    trisicell score ground.CFMatrix inferred.CFMatrix
    """

    tsc.settings.verbosity = "info"

    df_g = tsc.io.read(ground_file)
    df_s = tsc.io.read(inferred_file)

    ad = tsc.tl.ad(df_g, df_s)
    dl = tsc.tl.dl(df_g, df_s)
    mltd = tsc.tl.mltd(df_g, df_s)
    tpted = tsc.tl.tpted(df_g, df_s)

    tsc.logg.info(
        f"AD={ad:0.4f}\nDL={dl:0.4f}\nMLTSM={mltd[2]:0.4f}\nTPTED={tpted:0.4f}"
    )

    return None



import os

import click
import numpy as np
from joblib import Parallel, delayed

import trisicell as tsc


@click.command(short_help="Run SCITE.")
@click.argument(
    "genotype_file",
    required=True,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True
    ),
)
@click.argument(
    "alpha",
    required=True,
    type=float,
)
@click.argument(
    "beta",
    required=True,
    type=float,
)
@click.option(
    "--n_iters",
    "-l",
    default=1000000,
    type=int,
    show_default=True,
    help="Number of iterations.",
)
@click.option(
    "--n_restarts",
    "-r",
    default=3,
    type=int,
    show_default=True,
    help="Number of restarts.",
)
@click.option(
    "--experiment",
    "-e",
    is_flag=True,
    default=False,
    type=bool,
    show_default=True,
    help="Is in experiment mode.",
)
@click.option(
    "--output_file",
    "-o",
    default=None,
    type=click.Path(exists=False),
    show_default=True,
    help="Output file.",
)
def scite(genotype_file, alpha, beta, n_iters, n_restarts, experiment, output_file):
    """Tree inference for single-cell data :cite:`SCITE`.

    trisicell scite input.SC 0.0001 0.1 -l 1000000 -r 3 -e
    """

    if not output_file:
        outfile = os.path.splitext(genotype_file)[0]
    else:
        dirname, basename = tsc.ul.dir_base(output_file)
        tsc.ul.mkdir(dirname)

    tsc.settings.verbosity = "info"

    df_in = tsc.io.read(genotype_file)
    if experiment == False:
        tsc.settings.logfile = f"{outfile}.scite.log"
        df_out = tsc.tl.scite(
            df_in,
            alpha=alpha,
            beta=beta,
            n_iters=n_iters,
            n_restarts=n_restarts,
            save_inter=False,
            dir_inter=".",
            base_inter=None,
        )
        tsc.io.write(df_out, f"{outfile}.scite.CFMatrix")
    else:
        tsc.settings.logfile = f"{outfile}.scite.log"
        df_out, running_time, _, _ = tsc.tl.scite(
            df_in,
            alpha=alpha,
            beta=beta,
            n_iters=30000,
            n_restarts=1,
            save_inter=False,
            dir_inter=".",
            base_inter=None,
            experiment=True,
        )
        n_iters = int(2 * 30000 * 24 * 60 * 60 / running_time)

        dir_inter = tsc.ul.tmpdir(prefix="trisicell.", suffix=".scite", dirname=".")

        def run(i):
            do, r, s, b = tsc.tl.scite(
                df_in,
                alpha=alpha,
                beta=beta,
                n_iters=n_iters,
                n_restarts=1,
                save_inter=True,
                dir_inter=dir_inter,
                base_inter=f"{i}",
                experiment=True,
            )
            return do, r, s, b

        output = Parallel(n_jobs=3)(delayed(run)(i) for i in range(3))

        scores = [x[2] for x in output]
        betas = [x[3] for x in output]
        best_i = np.argmax(scores)
        df_out = output[best_i][0]

        tsc.ul.stat(df_in, df_out, alpha, beta, output[best_i][1])
        tsc.logg.info(f"score: {output[best_i][2]}")
        tsc.logg.info(f"beta: {output[best_i][3]}")
        tsc.logg.info(f"n_iters: {n_iters}")
        tsc.logg.info(f"scores: {','.join(list(map(str, scores)))}")
        tsc.logg.info(f"betas: {','.join(list(map(str, betas)))}")
        tsc.logg.info(f"picked: {best_i}")

        tsc.io.write(df_out, f"{outfile}.scite.CFMatrix")
        tsc.ul.cleanup(dir_inter)

    return None
